@@ -30,6 +30,7 @@
 // OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #endregion
 
+using Google.Protobuf.Collections;
 using Google.Protobuf.Compatibility;
 using System;
 
@@ -41,13 +42,14 @@ namespace Google.Protobuf.Reflection
     public sealed class FieldDescriptor : DescriptorBase, IComparable<FieldDescriptor>
     {
         private EnumDescriptor enumType;
+        private MessageDescriptor extendeeType;
         private MessageDescriptor messageType;
         private FieldType fieldType;
         private readonly string propertyName; // Annoyingly, needed in Crosslink.
         private IFieldAccessor accessor;
 
         /// <summary>
-        /// Get the field's containing message type.
+        /// Get the field's containing message type, or <c>null</c> if it is a field defined at the top level of a file as an extension.
         /// </summary>
         public MessageDescriptor ContainingType { get; }
 
@@ -64,8 +66,10 @@ namespace Google.Protobuf.Reflection
 
         internal FieldDescriptorProto Proto { get; }
 
+        internal Extension Extension { get; }
+
         internal FieldDescriptor(FieldDescriptorProto proto, FileDescriptor file,
-                                 MessageDescriptor parent, int index, string propertyName)
+                                 MessageDescriptor parent, int index, string propertyName, Extension extension)
             : base(file, file.ComputeFullName(parent, proto.Name), index)
         {
             Proto = proto;
@@ -96,6 +100,7 @@ namespace Google.Protobuf.Reflection
             // We could trust the generated code and check whether the type of the property is
             // a MapField, but that feels a tad nasty.
             this.propertyName = propertyName;
+            Extension = extension;
             JsonName =  Proto.JsonName == "" ? JsonFormatter.ToJsonName(Proto.Name) : Proto.JsonName;
         }
 
@@ -254,10 +259,45 @@ namespace Google.Protobuf.Reflection
             }
         }
 
+        /// <summary>
+        /// For extension fields, returns the extended type
+        /// </summary>
+        public MessageDescriptor ExtendeeType
+        {
+            get
+            {
+                if (!Proto.HasExtendee)
+                {
+                    throw new InvalidOperationException("ExtendeeType is only valid for extension fields.");
+                }
+                return extendeeType;
+            }
+        }
+
         /// <summary>
         /// The (possibly empty) set of custom options for this field.
         /// </summary>
-        public CustomOptions CustomOptions => Proto.Options?.CustomOptions ?? CustomOptions.Empty;
+        //[Obsolete("CustomOptions are obsolete. Use GetOption")]
+        public CustomOptions CustomOptions => Proto.Options.CustomOptions;
+
+        /* // uncomment this in the full proto2 support PR
+        /// <summary>
+        /// Gets a single value enum option for this descriptor
+        /// </summary>
+        public T GetOption<T>(Extension<FieldOptions, T> extension)
+        {
+            var value = Proto.Options.GetExtension(extension);
+            return value is IDeepCloneable<T> clonable ? clonable.Clone() : value;
+        }
+
+        /// <summary>
+        /// Gets a repeated value enum option for this descriptor
+        /// </summary>
+        public RepeatedField<T> GetOption<T>(RepeatedExtension<FieldOptions, T> extension)
+        {
+            return Proto.Options.GetExtension(extension).Clone();
+        }
+        */
 
         /// <summary>
         /// Look up and cross-link all field types etc.
@@ -320,6 +360,11 @@ namespace Google.Protobuf.Reflection
                 }
             }
 
+            if (Proto.HasExtendee)
+            {
+                extendeeType = File.DescriptorPool.LookupSymbol(Proto.Extendee, this) as MessageDescriptor;
+            }
+
             // Note: no attempt to perform any default value parsing
 
             File.DescriptorPool.AddFieldByNumber(this);
@@ -340,6 +385,11 @@ namespace Google.Protobuf.Reflection
             {
                 return null;
             }
+
+            if (Extension != null)
+            {
+                return new ExtensionAccessor(this);
+            }
             var property = ContainingType.ClrType.GetProperty(propertyName);
             if (property == null)
             {
@@ -350,4 +400,5 @@ namespace Google.Protobuf.Reflection
                 : (IFieldAccessor) new SingleFieldAccessor(property, this);
         }
     }
-}
\ No newline at end of file
+}
+ 
\ No newline at end of file
