@@ -37,7 +37,7 @@
 #include "upb.h"
 
 #define PHP_PROTOBUF_EXTNAME "protobuf"
-#define PHP_PROTOBUF_VERSION "3.9.2"
+#define PHP_PROTOBUF_VERSION "3.10.1"
 
 #define MAX_LENGTH_OF_INT64 20
 #define SIZEOF_INT64 8
@@ -46,6 +46,15 @@
 #define ARRAY_SIZE(x) \
     ((sizeof(x)/sizeof(0[x])) / ((size_t)(!(sizeof(x) % sizeof(0[x])))))
 
+#define PHP_PROTO_UNUSED(var) (void)var
+/* PHP_PROTO_ASSERT(): in release mode, we use the expression without letting
+ * it be evaluated.  This prevents "unused variable" warnings. */
+#ifdef NDEBUG
+#define PHP_PROTO_ASSERT(expr) do {} while (false && (expr))
+#else
+#define PHP_PROTO_ASSERT(expr) assert(expr)
+#endif
+
 // -----------------------------------------------------------------------------
 // PHP7 Wrappers
 // ----------------------------------------------------------------------------
