@@ -28,7 +28,62 @@
 // (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 // OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
+//                The Zend Engine License, version 2.00
+// Copyright (c) 1999-2002 Zend Technologies Ltd. All rights reserved.
+// --------------------------------------------------------------------
+//
+// Redistribution and use in source and binary forms, with or without
+// modification, is permitted provided that the following conditions
+// are met:
+//
+//   1. Redistributions of source code must retain the above copyright
+//      notice, this list of conditions and the following disclaimer.
+//
+//   2. Redistributions in binary form must reproduce the above
+//      copyright notice, this list of conditions and the following
+//      disclaimer in the documentation and/or other materials provided
+//      with the distribution.
+//
+//   3. The names "Zend" and "Zend Engine" must not be used to endorse
+//      or promote products derived from this software without prior
+//      permission from Zend Technologies Ltd. For written permission,
+//      please contact license@zend.com.
+//
+//   4. Zend Technologies Ltd. may publish revised and/or new versions
+//      of the license from time to time. Each version will be given a
+//      distinguishing version number.
+//      Once covered code has been published under a particular version
+//      of the license, you may always continue to use it under the
+//      terms of that version. You may also choose to use such covered
+//      code under the terms of any subsequent version of the license
+//      published by Zend Technologies Ltd. No one other than Zend
+//      Technologies Ltd. has the right to modify the terms applicable
+//      to covered code created under this License.
+//
+//   5. Redistributions of any form whatsoever must retain the following
+//      acknowledgment:
+//      "This product includes the Zend Engine, freely available at
+//      http://www.zend.com"
+//
+//   6. All advertising materials mentioning features or use of this
+//      software must display the following acknowledgment:
+//      "The Zend Engine is freely available at http://www.zend.com"
+//
+// THIS SOFTWARE IS PROVIDED BY ZEND TECHNOLOGIES LTD. ``AS IS'' AND
+// ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
+// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+// PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ZEND
+// TECHNOLOGIES LTD.  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
+// USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
+// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
+// OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+// SUCH DAMAGE.
+
 #include <Zend/zend_operators.h>
+#include <Zend/zend_exceptions.h>
 
 #include "protobuf.h"
 #include "utf8.h"
@@ -51,6 +106,13 @@ ZEND_BEGIN_ARG_INFO_EX(arg_check_repeated, 0, 0, 2)
   ZEND_ARG_INFO(0, klass)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_INFO_EX(arg_check_map, 0, 0, 3)
+  ZEND_ARG_INFO(1, val)
+  ZEND_ARG_INFO(0, key_type)
+  ZEND_ARG_INFO(0, value_type)
+  ZEND_ARG_INFO(0, klass)
+ZEND_END_ARG_INFO()
+
 static zend_function_entry util_methods[] = {
   PHP_ME(Util, checkInt32,  arg_check_optional, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
   PHP_ME(Util, checkUint32, arg_check_optional, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
@@ -63,6 +125,7 @@ static zend_function_entry util_methods[] = {
   PHP_ME(Util, checkString, arg_check_optional, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
   PHP_ME(Util, checkBytes,  arg_check_optional, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
   PHP_ME(Util, checkMessage, arg_check_message, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
+  PHP_ME(Util, checkMapField,    arg_check_map, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
   PHP_ME(Util, checkRepeatedField, arg_check_repeated,
          ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
   ZEND_FE_END
@@ -79,7 +142,7 @@ void util_init(TSRMLS_D) {
 // Type checking/conversion.
 // -----------------------------------------------------------------------------
 
-// This is modified from is_numeric_string in zend_operators.h. The behavior of 
+// This is modified from is_numeric_string in zend_operators.h. The behavior of
 // this function is the same as is_numeric_string, except that this takes
 // int64_t as input instead of long.
 static zend_uchar convert_numeric_string(
@@ -93,7 +156,7 @@ static zend_uchar convert_numeric_string(
     return IS_NULL;
   }
 
-  while (*str == ' ' || *str == '\t' || *str == '\n' || 
+  while (*str == ' ' || *str == '\t' || *str == '\n' ||
          *str == '\r' || *str == '\v' || *str == '\f') {
     str++;
     length--;
@@ -214,7 +277,7 @@ process_double:
                                        type##_t* type##_value) {             \
     int64_t lval;                                                            \
     double dval;                                                             \
-                                                                             \
+    TSRMLS_FETCH();                                                          \
     switch (convert_numeric_string(val, len, &lval, &dval)) {                \
       case IS_DOUBLE: {                                                      \
         return convert_double_to_##type(dval, type##_value);                 \
@@ -223,13 +286,15 @@ process_double:
         return convert_int64_to_##type(lval, type##_value);                  \
       }                                                                      \
       default:                                                               \
-        zend_error(E_USER_ERROR,                                             \
-                   "Given string value cannot be converted to integer.");    \
+        zend_throw_exception(NULL,                                           \
+                   "Given string value cannot be converted to integer.",     \
+                   0 TSRMLS_CC);                                             \
         return false;                                                        \
     }                                                                        \
   }                                                                          \
                                                                              \
   bool protobuf_convert_to_##type(zval* from, type##_t* to) {                \
+    TSRMLS_FETCH();                                                          \
     switch (Z_TYPE_P(from)) {                                                \
       case IS_LONG: {                                                        \
         return convert_int64_to_##type(Z_LVAL_P(from), to);                  \
@@ -242,8 +307,9 @@ process_double:
                                         to);                                 \
       }                                                                      \
       default: {                                                             \
-        zend_error(E_USER_ERROR,                                             \
-                   "Given value cannot be converted to integer.");           \
+        zend_throw_exception(NULL,                                           \
+                   "Given value cannot be converted to integer.",            \
+                   0 TSRMLS_CC);                                             \
         return false;                                                        \
       }                                                                      \
     }                                                                        \
@@ -273,6 +339,7 @@ CONVERT_TO_INTEGER(uint64);
     int64_t lval;                                                           \
     double dval;                                                            \
                                                                             \
+    TSRMLS_FETCH();                                                         \
     switch (convert_numeric_string(val, len, &lval, &dval)) {               \
       case IS_DOUBLE: {                                                     \
         *type##_value = (type)dval;                                         \
@@ -283,13 +350,15 @@ CONVERT_TO_INTEGER(uint64);
         return true;                                                        \
       }                                                                     \
       default:                                                              \
-        zend_error(E_USER_ERROR,                                            \
-                   "Given string value cannot be converted to integer.");   \
+        zend_throw_exception(NULL,                                          \
+                   "Given string value cannot be converted to integer.",    \
+                   0 TSRMLS_CC);                                            \
         return false;                                                       \
     }                                                                       \
   }                                                                         \
                                                                             \
   bool protobuf_convert_to_##type(zval* from, type* to) {                   \
+    TSRMLS_FETCH();                                                         \
     switch (Z_TYPE_P(from)) {                                               \
       case IS_LONG: {                                                       \
         return convert_int64_to_##type(Z_LVAL_P(from), to);                 \
@@ -302,8 +371,9 @@ CONVERT_TO_INTEGER(uint64);
                                         to);                                \
       }                                                                     \
       default: {                                                            \
-        zend_error(E_USER_ERROR,                                            \
-                   "Given value cannot be converted to integer.");          \
+        zend_throw_exception(NULL,                                          \
+                   "Given value cannot be converted to integer.",           \
+                   0 TSRMLS_CC);                                            \
         return false;                                                       \
       }                                                                     \
     }                                                                       \
@@ -316,10 +386,20 @@ CONVERT_TO_FLOAT(double);
 #undef CONVERT_TO_FLOAT
 
 bool protobuf_convert_to_bool(zval* from, int8_t* to) {
+  TSRMLS_FETCH();
   switch (Z_TYPE_P(from)) {
+#if PHP_MAJOR_VERSION < 7
     case IS_BOOL:
       *to = (int8_t)Z_BVAL_P(from);
       break;
+#else
+    case IS_TRUE:
+      *to = 1;
+      break;
+    case IS_FALSE:
+      *to = 0;
+      break;
+#endif
     case IS_LONG:
       *to = (int8_t)(Z_LVAL_P(from) != 0);
       break;
@@ -337,7 +417,9 @@ bool protobuf_convert_to_bool(zval* from, int8_t* to) {
       }
     } break;
     default: {
-      zend_error(E_USER_ERROR, "Given value cannot be converted to bool.");
+      zend_throw_exception(
+          NULL, "Given value cannot be converted to bool.",
+          0 TSRMLS_CC);
       return false;
     }
   }
@@ -345,21 +427,33 @@ bool protobuf_convert_to_bool(zval* from, int8_t* to) {
 }
 
 bool protobuf_convert_to_string(zval* from) {
+#if PHP_MAJOR_VERSION >= 7
+  if (Z_ISREF_P(from)) {
+    ZVAL_DEREF(from);
+  }
+#endif
+  TSRMLS_FETCH();
   switch (Z_TYPE_P(from)) {
     case IS_STRING: {
       return true;
     }
+#if PHP_MAJOR_VERSION < 7
     case IS_BOOL:
+#else
+    case IS_TRUE:
+    case IS_FALSE:
+#endif
     case IS_LONG:
     case IS_DOUBLE: {
-      int use_copy;
       zval tmp;
-      zend_make_printable_zval(from, &tmp, &use_copy);
+      php_proto_zend_make_printable_zval(from, &tmp);
       ZVAL_COPY_VALUE(from, &tmp);
       return true;
     }
     default:
-      zend_error(E_USER_ERROR, "Given value cannot be converted to string.");
+      zend_throw_exception(
+          NULL, "Given value cannot be converted to string.",
+          0 TSRMLS_CC);
       return false;
   }
 }
@@ -400,31 +494,167 @@ PHP_METHOD(Util, checkMessage) {
     RETURN_NULL();
   }
   if (!instanceof_function(Z_OBJCE_P(val), klass TSRMLS_CC)) {
-    zend_error(E_USER_ERROR, "Given value is not an instance of %s.",
-               klass->name);
+    zend_throw_exception_ex(NULL, 0 TSRMLS_CC,
+                            "Given value is not an instance of %s.",
+                            klass->name);
     return;
   }
   RETURN_ZVAL(val, 1, 0);
 }
 
-PHP_METHOD(Util, checkRepeatedField) {
-  zval* val;
-  long type;
-  const zend_class_entry* klass = NULL;
-  if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Ol|C", &val,
-                            repeated_field_type, &type, &klass) == FAILURE) {
-    return;
+void check_repeated_field(const zend_class_entry* klass, PHP_PROTO_LONG type,
+                          zval* val, zval* return_value) {
+#if PHP_MAJOR_VERSION >= 7
+  if (Z_ISREF_P(val)) {
+    ZVAL_DEREF(val);
   }
+#endif
 
-  RepeatedField *intern =
-      (RepeatedField *)zend_object_store_get_object(val TSRMLS_CC);
-  if (to_fieldtype(type) != intern->type) {
-    zend_error(E_USER_ERROR, "Incorrect repeated field type.");
-    return;
-  }
-  if (klass != NULL && intern->msg_ce != klass) {
-    zend_error(E_USER_ERROR, "Expect a repeated field of %s, but %s is given.",
-               klass->name, intern->msg_ce->name);
+  TSRMLS_FETCH();
+  if (Z_TYPE_P(val) == IS_ARRAY) {
+    HashTable* table = HASH_OF(val);
+    HashPosition pointer;
+    void* memory;
+
+#if PHP_MAJOR_VERSION < 7
+    zval* repeated_field;
+    MAKE_STD_ZVAL(repeated_field);
+#else
+    zval repeated_field;
+#endif
+
+    repeated_field_create_with_type(repeated_field_type, to_fieldtype(type),
+                                    klass, &repeated_field TSRMLS_CC);
+
+    for (zend_hash_internal_pointer_reset_ex(table, &pointer);
+         php_proto_zend_hash_get_current_data_ex(table, (void**)&memory,
+                                                 &pointer) == SUCCESS;
+         zend_hash_move_forward_ex(table, &pointer)) {
+      repeated_field_handlers->write_dimension(
+          CACHED_TO_ZVAL_PTR(repeated_field), NULL,
+          CACHED_PTR_TO_ZVAL_PTR((CACHED_VALUE*)memory) TSRMLS_CC);
+    }
+
+    RETURN_ZVAL(CACHED_TO_ZVAL_PTR(repeated_field), 1, 1);
+
+  } else if (Z_TYPE_P(val) == IS_OBJECT) {
+    if (!instanceof_function(Z_OBJCE_P(val), repeated_field_type TSRMLS_CC)) {
+      zend_throw_exception_ex(NULL, 0 TSRMLS_CC,
+                              "Given value is not an instance of %s.",
+                              repeated_field_type->name);
+      return;
+    }
+    RepeatedField* intern = UNBOX(RepeatedField, val);
+    if (to_fieldtype(type) != intern->type) {
+      zend_throw_exception_ex(NULL, 0 TSRMLS_CC,
+                              "Incorrect repeated field type.");
+      return;
+    }
+    if (klass != NULL && intern->msg_ce != klass) {
+      zend_throw_exception_ex(NULL, 0 TSRMLS_CC,
+                              "Expect a repeated field of %s, but %s is given.",
+                              klass->name, intern->msg_ce->name);
+      return;
+    }
+    RETURN_ZVAL(val, 1, 0);
+  } else {
+    zend_throw_exception_ex(NULL, 0 TSRMLS_CC,
+                            "Incorrect repeated field type.");
     return;
   }
 }
+
+PHP_METHOD(Util, checkRepeatedField) {
+  zval* val;
+  PHP_PROTO_LONG type;
+  const zend_class_entry* klass = NULL;
+  if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zl|C", &val, &type,
+                            &klass) == FAILURE) {
+    return;
+  }
+  RETURN_ZVAL(val, 1, 0);
+}
+
+void check_map_field(const zend_class_entry* klass, PHP_PROTO_LONG key_type,
+                     PHP_PROTO_LONG value_type, zval* val, zval* return_value) {
+#if PHP_MAJOR_VERSION >= 7
+  if (Z_ISREF_P(val)) {
+    ZVAL_DEREF(val);
+  }
+#endif
+
+  TSRMLS_FETCH();
+  if (Z_TYPE_P(val) == IS_ARRAY) {
+    HashTable* table = Z_ARRVAL_P(val);
+    HashPosition pointer;
+    zval key;
+    void* value;
+
+#if PHP_MAJOR_VERSION < 7
+    zval* map_field;
+    MAKE_STD_ZVAL(map_field);
+#else
+    zval map_field;
+#endif
+
+    map_field_create_with_type(map_field_type, to_fieldtype(key_type),
+                               to_fieldtype(value_type), klass,
+                               &map_field TSRMLS_CC);
+
+    for (zend_hash_internal_pointer_reset_ex(table, &pointer);
+         php_proto_zend_hash_get_current_data_ex(table, (void**)&value,
+                                                 &pointer) == SUCCESS;
+         zend_hash_move_forward_ex(table, &pointer)) {
+      zend_hash_get_current_key_zval_ex(table, &key, &pointer);
+      map_field_handlers->write_dimension(
+          CACHED_TO_ZVAL_PTR(map_field), &key,
+          CACHED_PTR_TO_ZVAL_PTR((CACHED_VALUE*)value) TSRMLS_CC);
+      zval_dtor(&key);
+    }
+
+    RETURN_ZVAL(CACHED_TO_ZVAL_PTR(map_field), 1, 1);
+  } else if (Z_TYPE_P(val) == IS_OBJECT) {
+    if (!instanceof_function(Z_OBJCE_P(val), map_field_type TSRMLS_CC)) {
+      zend_throw_exception_ex(NULL, 0 TSRMLS_CC,
+                              "Given value is not an instance of %s.",
+                              map_field_type->name);
+      return;
+    }
+    Map* intern = UNBOX(Map, val);
+    if (to_fieldtype(key_type) != intern->key_type) {
+      zend_throw_exception(
+          NULL, "Incorrect map field key type.",
+          0 TSRMLS_CC);
+      return;
+    }
+    if (to_fieldtype(value_type) != intern->value_type) {
+      zend_throw_exception(
+          NULL, "Incorrect map field value type.",
+          0 TSRMLS_CC);
+      return;
+    }
+    if (klass != NULL && intern->msg_ce != klass) {
+      zend_throw_exception_ex(NULL, 0 TSRMLS_CC,
+                              "Expect a map field of %s, but %s is given.",
+                              klass->name, intern->msg_ce->name);
+      return;
+    }
+    RETURN_ZVAL(val, 1, 0);
+  } else {
+      zend_throw_exception(
+          NULL, "Incorrect map field type.",
+          0 TSRMLS_CC);
+    return;
+  }
+}
+
+PHP_METHOD(Util, checkMapField) {
+  zval* val;
+  PHP_PROTO_LONG key_type, value_type;
+  const zend_class_entry* klass = NULL;
+  if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zll|C", &val, &key_type,
+                            &value_type, &klass) == FAILURE) {
+    return;
+  }
+  RETURN_ZVAL(val, 1, 0);
+}
